@@ -22,7 +22,6 @@ import java.io.BufferedReader;
 import java.io.BufferedWriter;
 import java.io.ByteArrayOutputStream;
 import java.io.File;
-import java.io.FileInputStream;
 import java.io.FileOutputStream;
 import java.io.IOException;
 import java.io.InputStream;
@@ -35,12 +34,15 @@ import java.net.URL;
 import java.net.URLStreamHandler;
 import java.security.CodeSigner;
 import java.security.CodeSource;
+import java.security.PrivilegedActionException;
+import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collection;
 import java.util.Enumeration;
 import java.util.HashMap;
 import java.util.HashSet;
 import java.util.Iterator;
+import java.util.List;
 import java.util.Map;
 import java.util.NoSuchElementException;
 import java.util.TreeSet;
@@ -50,6 +52,8 @@ import java.util.jar.Manifest;
 import java.util.zip.ZipEntry;
 import java.util.zip.ZipOutputStream;
 
+import static java.security.AccessController.doPrivileged;
+
 /**
  *
  * @author <a href="mailto:david.lloyd@redhat.com">David M. Lloyd</a>
@@ -64,6 +68,7 @@ final class JarFileResourceLoader extends AbstractResourceLoader implements Iter
     private final URL rootUrl;
     private final String relativePath;
     private final File fileOfJar;
+    private volatile List<String> directory;
 
     // protected by {@code this}
     private final Map<CodeSigners, CodeSource> codeSources = new HashMap<>();
@@ -128,8 +133,7 @@ final class JarFileResourceLoader extends AbstractResourceLoader implements Iter
             return null;
         }
         final long size = entry.getSize();
-        final InputStream is = jarFile.getInputStream(entry);
-        try {
+        try (final InputStream is = jarFile.getInputStream(entry)) {
             if (size == -1) {
                 // size unknown
                 final ByteArrayOutputStream baos = new ByteArrayOutputStream();
@@ -165,8 +169,6 @@ final class JarFileResourceLoader extends AbstractResourceLoader implements Iter
             } else {
                 throw new IOException("Resource is too large to be a valid class file");
             }
-        } finally {
-            StreamUtil.safeClose(is);
         }
     }
 
@@ -194,11 +196,8 @@ final class JarFileResourceLoader extends AbstractResourceLoader implements Iter
             if (jarEntry == null) {
                 manifest = null;
             } else {
-                InputStream inputStream = jarFile.getInputStream(jarEntry);
-                try {
+                try (final InputStream inputStream = jarFile.getInputStream(jarEntry)) {
                     manifest = new Manifest(inputStream);
-                } finally {
-                    StreamUtil.safeClose(inputStream);
                 }
             }
         }
@@ -239,7 +238,13 @@ final class JarFileResourceLoader extends AbstractResourceLoader implements Iter
             } catch (URISyntaxException x) {
                 throw new IllegalStateException(x);
             }
-            return new JarEntryResource(jarFile, entry, relativePath, new URL(null, getJarURI(uri, entry.getName()).toString(), (URLStreamHandler) null));
+            final URL url = new URL(null, getJarURI(uri, entry.getName()).toString(), (URLStreamHandler) null);
+            try {
+                doPrivileged(new GetURLConnectionAction(url));
+            } catch (PrivilegedActionException e) {
+                // ignored; the user might not even ask for the URL
+            }
+            return new JarEntryResource(jarFile, entry.getName(), relativePath, url);
         } catch (MalformedURLException e) {
             // must be invalid...?  (todo: check this out)
             return null;
@@ -250,26 +255,39 @@ final class JarFileResourceLoader extends AbstractResourceLoader implements Iter
     }
 
     public Iterator<Resource> iterateResources(String startPath, final boolean recursive) {
-        final JarFile jarFile = this.jarFile;
         if (relativePath != null) startPath = startPath.equals("") ? relativePath : relativePath + "/" + startPath;
         final String startName = PathUtils.canonicalize(PathUtils.relativize(startPath));
-        final Enumeration<JarEntry> entries = jarFile.entries();
+        List<String> directory = this.directory;
+        if (directory == null) {
+            synchronized (jarFile) {
+                directory = this.directory;
+                if (directory == null) {
+                    directory = new ArrayList<>();
+                    final Enumeration<JarEntry> entries = jarFile.entries();
+                    while (entries.hasMoreElements()) {
+                        final JarEntry jarEntry = entries.nextElement();
+                        if (! jarEntry.isDirectory()) {
+                            directory.add(jarEntry.getName());
+                        }
+                    }
+                    this.directory = directory;
+                }
+            }
+        }
+        final Iterator<String> iterator = directory.iterator();
         return new Iterator<Resource>() {
             private Resource next;
 
             public boolean hasNext() {
                 while (next == null) {
-                    if (! entries.hasMoreElements()) {
+                    if (! iterator.hasNext()) {
                         return false;
                     }
-                    final JarEntry entry = entries.nextElement();
-                    final String name = entry.getName();
+                    final String name = iterator.next();
                     if ((recursive ? PathUtils.isChild(startName, name) : PathUtils.isDirectChild(startName, name))) {
-                        if (!entry.isDirectory()) {
-                            try {
-                                next = new JarEntryResource(jarFile, entry, relativePath, getJarURI(new File(jarFile.getName()).toURI(), entry.getName()).toURL());
-                            } catch (Exception ignored) {
-                            }
+                        try {
+                            next = new JarEntryResource(jarFile, name, relativePath, getJarURI(new File(jarFile.getName()).toURI(), name).toURL());
+                        } catch (Exception ignored) {
                         }
                     }
                 }
@@ -299,19 +317,6 @@ final class JarFileResourceLoader extends AbstractResourceLoader implements Iter
         String relativePath = this.relativePath;
         // First check for an external index
         final JarFile jarFile = this.jarFile;
-        final String jarFileName = jarFile.getName();
-        final long jarModified = fileOfJar.lastModified();
-        final File indexFile = new File(jarFileName + ".index");
-        if (ResourceLoaders.USE_INDEXES) {
-            if (indexFile.exists()) {
-                final long indexModified = indexFile.lastModified();
-                if (indexModified != 0L && jarModified != 0L && indexModified >= jarModified) try {
-                    return readIndex(new FileInputStream(indexFile), index, relativePath);
-                } catch (IOException e) {
-                    index.clear();
-                }
-            }
-        }
         // Next check for an internal index
         JarEntry listEntry = jarFile.getJarEntry(INDEX_FILE);
         if (listEntry != null) {
@@ -323,10 +328,6 @@ final class JarFileResourceLoader extends AbstractResourceLoader implements Iter
         }
         // Next just read the JAR
         extractJarPaths(jarFile, relativePath, index);
-
-        if (ResourceLoaders.WRITE_INDEXES && relativePath == null) {
-            writeExternalIndex(indexFile, index);
-        }
         return index;
     }
 
@@ -351,6 +352,12 @@ final class JarFileResourceLoader extends AbstractResourceLoader implements Iter
         }
     }
 
+    public ResourceLoader createSubloader(final String relativePath, final String rootName) {
+        final String ourRelativePath = this.relativePath;
+        final String fixedPath = PathUtils.relativize(PathUtils.canonicalize(relativePath));
+        return new JarFileResourceLoader(rootName, jarFile, ourRelativePath == null ? fixedPath : ourRelativePath + "/" + fixedPath);
+    }
+
     static void extractJarPaths(final JarFile jarFile, String relativePath,
             final Collection<String> index) {
         index.add("");
@@ -380,16 +387,13 @@ final class JarFileResourceLoader extends AbstractResourceLoader implements Iter
         // Now try to write it
         boolean ok = false;
         try {
-            final BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(indexFile)));
-            try {
+            try (final BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(indexFile)))) {
                 for (String name : index) {
                     writer.write(name);
                     writer.write('\n');
                 }
                 writer.close();
                 ok = true;
-            } finally {
-                StreamUtil.safeClose(writer);
             }
         } catch (IOException e) {
             // failed, ignore
@@ -423,15 +427,13 @@ final class JarFileResourceLoader extends AbstractResourceLoader implements Iter
     }
 
     static void addInternalIndex(File file, boolean modify) throws IOException {
-        final JarFile oldJarFile = new JarFile(file, false);
-        try {
+        try (final JarFile oldJarFile = JDKSpecific.getJarFile(file, false)) {
             final Collection<String> index = new TreeSet<String>();
             final File outputFile;
 
             outputFile = new File(file.getAbsolutePath().replace(".jar", "-indexed.jar"));
 
-            final ZipOutputStream zo = new ZipOutputStream(new FileOutputStream(outputFile));
-            try {
+            try (final ZipOutputStream zo = new ZipOutputStream(new FileOutputStream(outputFile))) {
                 Enumeration<JarEntry> entries = oldJarFile.entries();
                 while (entries.hasMoreElements()) {
                     final JarEntry entry = entries.nextElement();
@@ -443,7 +445,7 @@ final class JarFileResourceLoader extends AbstractResourceLoader implements Iter
                         if (clone.getMethod() != ZipEntry.STORED)
                             clone.setCompressedSize(-1);
                         zo.putNextEntry(clone);
-                        StreamUtil.copy(oldJarFile.getInputStream(entry), zo);
+                        Utils.copy(oldJarFile.getInputStream(entry), zo);
                     }
 
                     // add to the index
@@ -460,15 +462,11 @@ final class JarFileResourceLoader extends AbstractResourceLoader implements Iter
 
                 // write index
                 zo.putNextEntry(new ZipEntry(INDEX_FILE));
-                final BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(zo));
-                try {
-                    for (String name : index) {
+                try (final BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(zo))) {
+                    for (final String name : index) {
                         writer.write(name);
                         writer.write('\n');
                     }
-                    writer.close();
-                } finally {
-                    StreamUtil.safeClose(writer);
                 }
                 zo.close();
                 oldJarFile.close();
@@ -479,11 +477,7 @@ final class JarFileResourceLoader extends AbstractResourceLoader implements Iter
                         throw new IOException("failed to rename " + outputFile.getAbsolutePath() + " to " + file.getAbsolutePath());
                     }
                 }
-            } finally {
-                StreamUtil.safeClose(zo);
             }
-        } finally {
-            StreamUtil.safeClose(oldJarFile);
         }
     }
 
