@@ -20,18 +20,22 @@ package org.jboss.modules;
 
 import java.io.IOException;
 import java.io.InputStream;
+import java.lang.invoke.MethodHandle;
+import java.lang.invoke.MethodHandles;
+import java.lang.invoke.MethodType;
 import java.lang.reflect.InvocationTargetException;
-import java.lang.reflect.Method;
-import java.lang.reflect.Modifier;
 import java.net.URL;
 import java.net.URLConnection;
 import java.security.AccessController;
+import java.security.Permission;
 import java.security.PermissionCollection;
+import java.security.Permissions;
 import java.security.PrivilegedAction;
 import java.util.ArrayList;
 import java.util.Collections;
 import java.util.Enumeration;
 import java.util.HashMap;
+import java.util.HashSet;
 import java.util.Iterator;
 import java.util.LinkedHashMap;
 import java.util.List;
@@ -50,7 +54,6 @@ import org.jboss.modules.filter.PathFilters;
 import org.jboss.modules.log.ModuleLogger;
 import org.jboss.modules.log.NoopModuleLogger;
 
-import __redirected.__JAXPRedirected;
 import org.jboss.modules.security.ModularPermissionFactory;
 
 /**
@@ -69,6 +72,7 @@ import org.jboss.modules.security.ModularPermissionFactory;
 public final class Module {
 
     private static final AtomicReference<ModuleLoader> BOOT_MODULE_LOADER;
+    private static final MethodType MAIN_METHOD_TYPE = MethodType.methodType(void.class, String[].class);
 
     static {
         log = NoopModuleLogger.getInstance();
@@ -84,9 +88,7 @@ public final class Module {
 
         final String pkgsString = AccessController.doPrivileged(new PropertyReadAction("jboss.modules.system.pkgs"));
         final List<String> list = new ArrayList<String>();
-        list.add("java.");
-        list.add("sun.reflect.");
-        list.add("__redirected.");
+        JDKSpecific.addInternalPackages(list);
         if (pkgsString != null) {
             int i;
             int nc = -1;
@@ -104,14 +106,15 @@ public final class Module {
                 }
             } while (nc != -1);
         }
-        systemPackages = list.toArray(list.toArray(new String[list.size()]));
+        final String[] noStrings = new String[0];
+        systemPackages = list.toArray(noStrings);
         final ListIterator<String> iterator = list.listIterator();
         // http://youtrack.jetbrains.net/issue/IDEA-72097
         //noinspection WhileLoopReplaceableByForEach
         while (iterator.hasNext()) {
             iterator.set(iterator.next().replace('.', '/'));
         }
-        systemPaths = list.toArray(list.toArray(new String[list.size()]));
+        systemPaths = list.toArray(noStrings);
 
         AccessController.doPrivileged(new PrivilegedAction<Void>() {
             public Void run() {
@@ -125,9 +128,6 @@ public final class Module {
                 } catch (Throwable t) {
                     // todo log a warning or something
                 }
-
-                __JAXPRedirected.initAll();
-
                 return null;
             }
         });
@@ -150,7 +150,7 @@ public final class Module {
      * @throws SecurityException always
      */
     public static ModulesPrivateAccess getPrivateAccess() {
-        if (CallerContext.getCallingClass() == ModularPermissionFactory.class) {
+        if (JDKSpecific.getCallingClass() == ModularPermissionFactory.class) {
             return PRIVATE_ACCESS;
         }
         throw new SecurityException();
@@ -167,9 +167,9 @@ public final class Module {
     // immutable properties
 
     /**
-     * The identifier of this module.
+     * The name of this module.
      */
-    private final ModuleIdentifier identifier;
+    private final String name;
     /**
      * The name of the main class, if any (may be {@code null}).
      */
@@ -194,8 +194,16 @@ public final class Module {
      * The assigned permission collection.
      */
     private final PermissionCollection permissionCollection;
+    /**
+     * The (optional) module version.
+     */
+    private final Version version;
 
     // mutable properties
+    /**
+     * Module aliases
+     */
+    volatile Set<String> aliases = new HashSet<>();
 
     /**
      * The linkage state.
@@ -211,6 +219,8 @@ public final class Module {
     private static final RuntimePermission ADD_CONTENT_HANDLER_FACTORY;
     private static final RuntimePermission ADD_URL_STREAM_HANDLER_FACTORY;
 
+    private static final PermissionCollection NO_PERMISSIONS = noPermissions();
+
     /**
      * Construct a new instance from a module specification.
      *
@@ -221,19 +231,37 @@ public final class Module {
         this.moduleLoader = moduleLoader;
 
         // Initialize state from the spec.
-        identifier = spec.getModuleIdentifier();
+        name = spec.getName();
         mainClassName = spec.getMainClass();
         fallbackLoader = spec.getFallbackLoader();
-        permissionCollection = spec.getPermissionCollection();
+        final PermissionCollection permissionCollection = spec.getPermissionCollection();
+        this.permissionCollection = permissionCollection == null ? NO_PERMISSIONS : permissionCollection.isReadOnly() ? permissionCollection : copyPermissions(permissionCollection);
         //noinspection ThisEscapedInObjectConstruction
         final ModuleClassLoader.Configuration configuration = new ModuleClassLoader.Configuration(this, spec.getAssertionSetting(), spec.getResourceLoaders(), spec.getClassFileTransformer());
         final ModuleClassLoaderFactory factory = spec.getModuleClassLoaderFactory();
+        final Map<String, String> properties = spec.getProperties();
+        this.properties = properties.isEmpty() ? Collections.<String, String>emptyMap() : new LinkedHashMap<String, String>(properties);
+        this.version = spec.getVersion();
         ModuleClassLoader moduleClassLoader = null;
         if (factory != null) moduleClassLoader = factory.create(configuration);
         if (moduleClassLoader == null) moduleClassLoader = new ModuleClassLoader(configuration);
         this.moduleClassLoader = moduleClassLoader;
-        final Map<String, String> properties = spec.getProperties();
-        this.properties = properties.isEmpty() ? Collections.<String, String>emptyMap() : new LinkedHashMap<String, String>(properties);
+    }
+
+    private static PermissionCollection noPermissions() {
+        final Permissions permissions = new Permissions();
+        permissions.setReadOnly();
+        return permissions;
+    }
+
+    private static PermissionCollection copyPermissions(PermissionCollection permissionCollection) {
+        final Permissions permissions = new Permissions();
+        final Enumeration<Permission> elements = permissionCollection.elements();
+        while (elements.hasMoreElements()) {
+            permissions.add(elements.nextElement());
+        }
+        permissions.setReadOnly();
+        return permissions;
     }
 
     LocalLoader getFallbackLoader() {
@@ -273,7 +301,9 @@ public final class Module {
      * @param rootPath the module root to search
      * @param resourcePath the path of the resource
      * @return the resource
+     * @deprecated Resource root names are deprecated.
      */
+    @Deprecated
     public Resource getExportedResource(final String rootPath, final String resourcePath) {
         return moduleClassLoader.loadResourceLocal(rootPath, resourcePath);
     }
@@ -287,31 +317,44 @@ public final class Module {
      * @throws ClassNotFoundException if the main class is not found
      */
     public void run(final String[] args) throws NoSuchMethodException, InvocationTargetException, ClassNotFoundException {
+        run(mainClassName, args);
+    }
+
+    /**
+     * Run the given main class in this module.
+     *
+     * @param className the class name to run (must not be {@code null})
+     * @param args the arguments to pass
+     * @throws NoSuchMethodException if there is no main method
+     * @throws InvocationTargetException if the main method failed
+     * @throws ClassNotFoundException if the main class is not found
+     */
+    public void run(final String className, final String[] args) throws NoSuchMethodException, InvocationTargetException, ClassNotFoundException {
+        if (className == null) {
+            throw new NoSuchMethodException("No main class defined for " + this);
+        }
+        final ClassLoader oldClassLoader = SecurityActions.setContextClassLoader(moduleClassLoader);
         try {
-            if (mainClassName == null) {
-                throw new NoSuchMethodException("No main class defined for " + this);
-            }
-            final ClassLoader oldClassLoader = SecurityActions.setContextClassLoader(moduleClassLoader);
+            final Class<?> mainClass = Class.forName(className, false, moduleClassLoader);
             try {
-                final Class<?> mainClass = Class.forName(mainClassName, false, moduleClassLoader);
-                try {
-                    Class.forName(mainClassName, true, moduleClassLoader);
-                } catch (Throwable t) {
-                    throw new InvocationTargetException(t, "Failed to initialize main class '" + mainClassName + "'");
-                }
-                final Method mainMethod = mainClass.getMethod("main", String[].class);
-                final int modifiers = mainMethod.getModifiers();
-                if (! Modifier.isStatic(modifiers)) {
-                    throw new NoSuchMethodException("Main method is not static for " + this);
-                }
-                // ignore the return value
-                mainMethod.invoke(null, new Object[] {args});
-            } finally {
-                SecurityActions.setContextClassLoader(oldClassLoader);
+                Class.forName(className, true, moduleClassLoader);
+            } catch (Throwable t) {
+                throw new InvocationTargetException(t, "Failed to initialize main class '" + className + "'");
             }
-        } catch (IllegalAccessException e) {
-            // unexpected; should be public
-            throw new IllegalAccessError(e.getMessage());
+            final MethodHandles.Lookup lookup = MethodHandles.lookup();
+            final MethodHandle methodHandle;
+            try {
+                methodHandle = lookup.findStatic(mainClass, "main", MAIN_METHOD_TYPE);
+            } catch (IllegalAccessException e) {
+                throw new NoSuchMethodException("The main method is not public");
+            }
+            try {
+                methodHandle.invokeExact(args);
+            } catch (Throwable throwable) {
+                throw new InvocationTargetException(throwable);
+            }
+        } finally {
+            SecurityActions.setContextClassLoader(oldClassLoader);
         }
     }
 
@@ -319,9 +362,20 @@ public final class Module {
      * Get this module's identifier.
      *
      * @return the identifier
+     * @deprecated Use {@link #getName()} instead.
      */
+    @Deprecated
     public ModuleIdentifier getIdentifier() {
-        return identifier;
+        return ModuleIdentifier.fromString(getName());
+    }
+
+    /**
+     * Get this module's name.
+     *
+     * @return this module's name
+     */
+    public String getName() {
+        return name;
     }
 
     /**
@@ -344,6 +398,30 @@ public final class Module {
         return ServiceLoader.load(serviceType, moduleClassLoader);
     }
 
+    /**
+     * Load a service loader from this module, without looking at dependencies.
+     *
+     * @param serviceType the service type class
+     * @param <S> the service type
+     * @return the service loader
+     */
+    public <S> ServiceLoader<S> loadServiceDirectly(Class<S> serviceType) {
+        return ServiceLoader.load(serviceType, new ClassLoader(null) {
+            public Enumeration<URL> getResources(final String name) throws IOException {
+                final Enumeration<Resource> resourceEnumeration = Collections.enumeration(getClassLoader().getLocalLoader().loadResourceLocal(name));
+                return new Enumeration<URL>() {
+                    public boolean hasMoreElements() {
+                        return resourceEnumeration.hasMoreElements();
+                    }
+
+                    public URL nextElement() {
+                        return resourceEnumeration.nextElement().getURL();
+                    }
+                };
+            }
+        });
+    }
+
     /**
      * Load a service loader from a module in the caller's module loader. The caller's
      * module loader refers to the loader of the module of the class that calls this method.
@@ -355,9 +433,27 @@ public final class Module {
      * @param serviceType the service type class
      * @return the loaded service from the caller's module
      * @throws ModuleLoadException if the named module failed to load
+     * @deprecated Use {@link #loadServiceFromCallerModuleLoader(String, Class)} instead.
      */
+    @Deprecated
     public static <S> ServiceLoader<S> loadServiceFromCallerModuleLoader(ModuleIdentifier identifier, Class<S> serviceType) throws ModuleLoadException {
-        return getCallerModuleLoader().loadModule(identifier).loadService(serviceType);
+        return loadServiceFromCallerModuleLoader(identifier.toString(), serviceType);
+    }
+
+    /**
+     * Load a service loader from a module in the caller's module loader. The caller's
+     * module loader refers to the loader of the module of the class that calls this method.
+     * Note that {@link #loadService(Class)} is more efficient since it does not need to crawl
+     * the stack.
+     *
+     * @param <S> the the service type
+     * @param name the module name containing the service loader
+     * @param serviceType the service type class
+     * @return the loaded service from the caller's module
+     * @throws ModuleLoadException if the named module failed to load
+     */
+    public static <S> ServiceLoader<S> loadServiceFromCallerModuleLoader(String name, Class<S> serviceType) throws ModuleLoadException {
+        return getCallerModuleLoader().loadModule(name).loadService(serviceType);
     }
 
     /**
@@ -441,6 +537,16 @@ public final class Module {
         return loader;
     }
 
+    /**
+     * Get the system module loader.  This is the module loader which contains the Java platform modules, plus
+     * a module for JBoss Modules itself, called {@code org.jboss.modules}.
+     *
+     * @return the system module loader
+     */
+    public static ModuleLoader getSystemModuleLoader() {
+        return Utils.JDK_MODULE_LOADER;
+    }
+
     static void initBootModuleLoader(ModuleLoader loader) {
         BOOT_MODULE_LOADER.set(loader);
     }
@@ -477,9 +583,24 @@ public final class Module {
      * @param identifier the module identifier
      * @return the module
      * @throws ModuleLoadException if the module could not be loaded
+     * @deprecated Use {@link #getModuleFromCallerModuleLoader(String)} instead.
      */
+    @Deprecated
     public static Module getModuleFromCallerModuleLoader(final ModuleIdentifier identifier) throws ModuleLoadException {
-        return getCallerModuleLoader().loadModule(identifier);
+        return getModuleFromCallerModuleLoader(identifier.toString());
+    }
+
+    /**
+     * Get a module from the current module loader. Note that this must crawl the
+     * stack to determine this, so other mechanisms are more efficient.
+     * @see #getCallerModuleLoader()
+     *
+     * @param name the module name
+     * @return the module
+     * @throws ModuleLoadException if the module could not be loaded
+     */
+    public static Module getModuleFromCallerModuleLoader(final String name) throws ModuleLoadException {
+        return getCallerModuleLoader().loadModule(name);
     }
 
     /**
@@ -490,7 +611,7 @@ public final class Module {
      * @return the current module
      */
     public static Module getCallerModule() {
-        return forClass(CallerContext.getCallingClass());
+        return forClass(JDKSpecific.getCallingUserClass());
     }
 
     /**
@@ -499,9 +620,22 @@ public final class Module {
      * @param identifier the module identifier
      * @return the module
      * @throws ModuleLoadException if an error occurs
+     * @deprecated Use {@link #getModule(String)} instead.
      */
+    @Deprecated
     public Module getModule(final ModuleIdentifier identifier) throws ModuleLoadException {
-        return moduleLoader.loadModule(identifier);
+        return getModule(identifier.toString());
+    }
+
+    /**
+     * Get the module with the given identifier from the module loader used by this module.
+     *
+     * @param name the module name
+     * @return the module
+     * @throws ModuleLoadException if an error occurs
+     */
+    public Module getModule(final String name) throws ModuleLoadException {
+        return moduleLoader.loadModule(name);
     }
 
     /**
@@ -515,10 +649,29 @@ public final class Module {
      * @return the class
      * @throws ModuleLoadException if the module could not be loaded
      * @throws ClassNotFoundException if the class could not be loaded
+     * @deprecated Use {@link #loadClassFromBootModuleLoader(String, String)} instead.
      */
+    @Deprecated
     public static Class<?> loadClassFromBootModuleLoader(final ModuleIdentifier moduleIdentifier, final String className)
             throws ModuleLoadException, ClassNotFoundException {
-        return Class.forName(className, true, getBootModuleLoader().loadModule(moduleIdentifier).getClassLoader());
+        return loadClassFromBootModuleLoader(moduleIdentifier.toString(), className);
+    }
+
+    /**
+     * Load a class from a module in the system module loader.
+     *
+     * @see #getBootModuleLoader()
+     *
+     * @param name the name of the module from which the class
+     *        should be loaded
+     * @param className the class name to load
+     * @return the class
+     * @throws ModuleLoadException if the module could not be loaded
+     * @throws ClassNotFoundException if the class could not be loaded
+     */
+    public static Class<?> loadClassFromBootModuleLoader(final String name, final String className)
+            throws ModuleLoadException, ClassNotFoundException {
+        return Class.forName(className, true, getBootModuleLoader().loadModule(name).getClassLoader());
     }
 
     /**
@@ -532,10 +685,28 @@ public final class Module {
      * @return the class
      * @throws ModuleLoadException if the module could not be loaded
      * @throws ClassNotFoundException if the class could not be loaded
+     * @deprecated Use {@link #loadClassFromCallerModuleLoader(String, String)} instead.
      */
     public static Class<?> loadClassFromCallerModuleLoader(final ModuleIdentifier moduleIdentifier, final String className)
             throws ModuleLoadException, ClassNotFoundException {
-        return Class.forName(className, true, getModuleFromCallerModuleLoader(moduleIdentifier).getClassLoader());
+        return loadClassFromCallerModuleLoader(moduleIdentifier.toString(), className);
+    }
+
+    /**
+     * Load a class from a module in the caller's module loader.
+     *
+     * @see #getCallerModuleLoader()
+     *
+     * @param name the name of the module from which the class
+     *        should be loaded
+     * @param className the class name to load
+     * @return the class
+     * @throws ModuleLoadException if the module could not be loaded
+     * @throws ClassNotFoundException if the class could not be loaded
+     */
+    public static Class<?> loadClassFromCallerModuleLoader(final String name, final String className)
+            throws ModuleLoadException, ClassNotFoundException {
+        return Class.forName(className, true, getModuleFromCallerModuleLoader(name).getClassLoader());
     }
 
     /**
@@ -585,23 +756,33 @@ public final class Module {
         }
         log.trace("Attempting to find resource %s in %s", canonPath, this);
         final String path = pathOf(canonPath);
+        final URLConnectionResource jaxpResource = ModuleClassLoader.jaxpImplResources.get(canonPath);
         final Map<String, List<LocalLoader>> paths = getPathsUnchecked();
         final List<LocalLoader> loaders = paths.get(path);
         if (loaders != null) {
             for (LocalLoader loader : loaders) {
-                final List<Resource> resourceList = loader.loadResourceLocal(canonPath);
-                for (Resource resource : resourceList) {
-                    return resource.getURL();
+                final Iterator<Resource> iterator = loader.loadResourceLocal(canonPath).iterator();
+                if (iterator.hasNext()) {
+                    final URL url = iterator.next().getURL();
+                    if (jaxpResource != null) log.jaxpResourceLoaded(url, this);
+                    return url;
                 }
             }
         }
         final LocalLoader fallbackLoader = this.fallbackLoader;
         if (fallbackLoader != null) {
-            final List<Resource> resourceList = fallbackLoader.loadResourceLocal(canonPath);
-            for (Resource resource : resourceList) {
-                return resource.getURL();
+            final Iterator<Resource> iterator = fallbackLoader.loadResourceLocal(canonPath).iterator();
+            if (iterator.hasNext()) {
+                final URL url = iterator.next().getURL();
+                if (jaxpResource != null) log.jaxpResourceLoaded(url, this);
+                return url;
             }
         }
+        if (jaxpResource != null) {
+            final URL url = jaxpResource.getURL();
+            log.jaxpResourceLoaded(url, this);
+            return url;
+        }
         return null;
     }
 
@@ -620,12 +801,16 @@ public final class Module {
         }
         log.trace("Attempting to find resource %s in %s", canonPath, this);
         final String path = pathOf(canonPath);
+        final URLConnectionResource jaxpResource = ModuleClassLoader.jaxpImplResources.get(canonPath);
         final Map<String, List<LocalLoader>> paths = getPathsUnchecked();
         final List<LocalLoader> loaders = paths.get(path);
         if (loaders != null) {
             for (LocalLoader loader : loaders) {
                 final List<Resource> resourceList = loader.loadResourceLocal(canonPath);
-                for (Resource resource : resourceList) {
+                final Iterator<Resource> iterator = resourceList.iterator();
+                if (iterator.hasNext()) {
+                    final Resource resource = iterator.next();
+                    if (jaxpResource != null) log.jaxpResourceLoaded(resource.getURL(), this);
                     return resource.openStream();
                 }
             }
@@ -633,10 +818,17 @@ public final class Module {
         final LocalLoader fallbackLoader = this.fallbackLoader;
         if (fallbackLoader != null) {
             final List<Resource> resourceList = fallbackLoader.loadResourceLocal(canonPath);
-            for (Resource resource : resourceList) {
+            final Iterator<Resource> iterator = resourceList.iterator();
+            if (iterator.hasNext()) {
+                final Resource resource = iterator.next();
+                if (jaxpResource != null) log.jaxpResourceLoaded(resource.getURL(), this);
                 return resource.openStream();
             }
         }
+        if (jaxpResource != null) {
+            log.jaxpResourceLoaded(jaxpResource.getURL(), this);
+            return jaxpResource.openStream();
+        }
         return null;
     }
 
@@ -659,6 +851,7 @@ public final class Module {
         }
         log.trace("Attempting to find all resources %s in %s", canonPath, this);
         final String path = pathOf(canonPath);
+        final URLConnectionResource jaxpResource = ModuleClassLoader.jaxpImplResources.get(canonPath);
         final Map<String, List<LocalLoader>> paths = getPathsUnchecked();
         final List<LocalLoader> loaders = paths.get(path);
 
@@ -667,7 +860,9 @@ public final class Module {
             for (LocalLoader loader : loaders) {
                 final List<Resource> resourceList = loader.loadResourceLocal(canonPath);
                 for (Resource resource : resourceList) {
-                    list.add(resource.getURL());
+                    final URL url = resource.getURL();
+                    if (jaxpResource != null) log.jaxpResourceLoaded(url, this);
+                    list.add(url);
                 }
             }
         }
@@ -675,9 +870,16 @@ public final class Module {
         if (fallbackLoader != null) {
             final List<Resource> resourceList = fallbackLoader.loadResourceLocal(canonPath);
             for (Resource resource : resourceList) {
-                list.add(resource.getURL());
+                final URL url = resource.getURL();
+                if (jaxpResource != null) log.jaxpResourceLoaded(url, this);
+                list.add(url);
             }
         }
+        if (jaxpResource != null) {
+            final URL url = jaxpResource.getURL();
+            log.jaxpResourceLoaded(url, this);
+            list.add(url);
+        }
 
         return list.size() == 0 ? ConcurrentClassLoader.EMPTY_ENUMERATION : Collections.enumeration(list);
     }
@@ -795,7 +997,7 @@ public final class Module {
      * @return the set of paths
      * @throws ModuleLoadException if the module was previously unlinked, and there was an exception while linking
      */
-    public final Set<String> getImportedPaths() throws ModuleLoadException {
+    public Set<String> getImportedPaths() throws ModuleLoadException {
         return Collections.unmodifiableSet(getPaths().keySet());
     }
 
@@ -877,6 +1079,15 @@ public final class Module {
         return new ArrayList<String>(properties.keySet());
     }
 
+    /**
+     * Get the module version.
+     *
+     * @return the module version, or {@code null} if none was set
+     */
+    public Version getVersion() {
+        return version;
+    }
+
     /**
      * Get the string representation of this module.
      *
@@ -884,7 +1095,15 @@ public final class Module {
      */
     @Override
     public String toString() {
-        return "Module \"" + identifier + "\"" + " from " + moduleLoader.toString();
+        StringBuilder b = new StringBuilder();
+        b.append("Module \"");
+        b.append(getName());
+        b.append("\"");
+        if (version != null) {
+            b.append(" version ").append(version);
+        }
+        b.append(" from ").append(moduleLoader);
+        return b.toString();
     }
 
     /**
@@ -978,7 +1197,12 @@ public final class Module {
         return NativeLibraryResourceLoader.getArchName();
     }
 
-    PermissionCollection getPermissionCollection() {
+    /**
+     * Get the module's configured permission collection.
+     *
+     * @return the module permission collection
+     */
+    public PermissionCollection getPermissionCollection() {
         return permissionCollection;
     }
 
@@ -1019,13 +1243,13 @@ public final class Module {
             if (dependency instanceof ModuleDependency) {
                 final ModuleDependency moduleDependency = (ModuleDependency) dependency;
                 final ModuleLoader moduleLoader = moduleDependency.getModuleLoader();
-                final ModuleIdentifier id = moduleDependency.getIdentifier();
+                final String name = moduleDependency.getName();
                 final Module module;
 
                 try {
                     long pauseStart = Metrics.getCurrentCPUTime();
                     try {
-                        module = moduleLoader.preloadModule(id);
+                        module = moduleLoader.preloadModule(name);
                     } finally {
                         subtract += Metrics.getCurrentCPUTime() - pauseStart;
                     }
@@ -1033,13 +1257,13 @@ public final class Module {
                     if (moduleDependency.isOptional()) {
                         continue;
                     } else {
-                        log.trace("Module %s, dependency %s preload failed: %s", getIdentifier(), moduleDependency.getIdentifier(), ex);
+                        log.trace("Module %s, dependency %s preload failed: %s", getIdentifier(), moduleDependency.getName(), ex);
                         throw ex;
                     }
                 }
                 if (module == null) {
                     if (!moduleDependency.isOptional()) {
-                        throw new ModuleNotFoundException(id.toString());
+                        throw new ModuleNotFoundException(name);
                     }
                     continue;
                 }
@@ -1170,13 +1394,13 @@ public final class Module {
                 if (dependency instanceof ModuleDependency) {
                     final ModuleDependency moduleDependency = (ModuleDependency) dependency;
                     final ModuleLoader moduleLoader = moduleDependency.getModuleLoader();
-                    final ModuleIdentifier id = moduleDependency.getIdentifier();
+                    final String name  = moduleDependency.getName();
                     final Module module;
 
                     try {
                         long pauseStart = Metrics.getCurrentCPUTime();
                         try {
-                            module = moduleLoader.preloadModule(id);
+                            module = moduleLoader.preloadModule(name);
                         } finally {
                             subtract += Metrics.getCurrentCPUTime() - pauseStart;
                         }
@@ -1184,13 +1408,13 @@ public final class Module {
                         if (moduleDependency.isOptional()) {
                             continue;
                         } else {
-                            log.trace("Module %s, dependency %s preload failed: %s", getIdentifier(), moduleDependency.getIdentifier(), ex);
+                            log.trace("Module %s, dependency %s preload failed: %s", getIdentifier(), moduleDependency.getName(), ex);
                             throw ex;
                         }
                     }
                     if (module == null) {
                         if (!moduleDependency.isOptional()) {
-                            throw new ModuleNotFoundException(id.toString());
+                            throw new ModuleNotFoundException(name);
                         }
                         continue;
                     }
