@@ -26,11 +26,27 @@ package org.jboss.modules;
 public interface ModuleFinder {
 
     /**
-     * Find a module specification for the given identifier.
+     * Find a module specification for the given name.  The default implementation delegates to the legacy
+     * {@link #findModule(ModuleIdentifier, ModuleLoader)} method.
      *
-     * @param identifier the module identifier
+     * @param name the module name
      * @param delegateLoader the module loader from which dependencies should be resolved
      * @return the module specification, or {@code null} if no specification is found for this identifier
      */
-    ModuleSpec findModule(ModuleIdentifier identifier, ModuleLoader delegateLoader) throws ModuleLoadException;
+    default ModuleSpec findModule(String name, ModuleLoader delegateLoader) throws ModuleLoadException {
+        return findModule(ModuleIdentifier.fromString(name), delegateLoader);
+    }
+
+    /**
+     * Find a module specification for the given identifier.  The default implementation returns {@code null}.  This
+     * method will never be called by the module system unless {@link #findModule(String, ModuleLoader)} is left
+     * unimplemented.
+     *
+     * @param moduleIdentifier the module identifier
+     * @param delegateLoader the module loader from which dependencies should be resolved
+     * @return the module specification, or {@code null} if no specification is found for this identifier
+     */
+    default ModuleSpec findModule(ModuleIdentifier moduleIdentifier, ModuleLoader delegateLoader) throws ModuleLoadException {
+        return null;
+    }
 }
