@@ -25,9 +25,12 @@ import java.lang.management.ManagementFactory;
 import java.lang.reflect.Method;
 import java.lang.reflect.Modifier;
 import java.lang.reflect.UndeclaredThrowableException;
+import java.net.URL;
+import java.security.CodeSource;
 import java.security.PrivilegedAction;
 import java.security.PrivilegedActionException;
 import java.security.PrivilegedExceptionAction;
+import java.security.ProtectionDomain;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collection;
@@ -1149,6 +1152,30 @@ public class ModuleLoader {
             return result;
         }
 
+        public String getClassLocation(final String moduleName, final String className) {
+            final ModuleLoader loader = getModuleLoader();
+            final Module module = loadModule(moduleName, loader);
+            final Class<?> clazz;
+            try {
+                clazz = Class.forName(className, false, module.getClassLoaderPrivate());
+            } catch (ClassNotFoundException e) {
+                return null;
+            }
+            final ProtectionDomain pd = clazz.getProtectionDomain();
+            if (pd == null) {
+                return null;
+            }
+            final CodeSource cs = pd.getCodeSource();
+            if (cs == null) {
+                return null;
+            }
+            final URL url = cs.getLocation();
+            if (url == null) {
+                return null;
+            }
+            return url.toString();
+        }
+
         private Module loadModule(final String name, final ModuleLoader loader) {
             try {
                 final Module module = loader.findLoadedModuleLocal(name);
